open import Agda.Primitive renaming (Set to Type)
open import Data.Bool
open import Data.Empty
open import Data.Unit
open import Data.Nat

import Data.String as S
import Data.List as L
open S using (String)
open L using (List)

module Base where

case_of_ : {A B : Type} → A → (A → B) → B
case x of f = f x

is-false : Bool → Type
is-false true = ⊥
is-false false = ⊤

record Monad (M : Type → Type) : Type₁ where
  field
    pure : {A : Type} → A → M A
    _>>=_ : {A B : Type} → M A → (A → M B) → M B

  _>>_ : {A B : Type} → M A → M B → M B
  a >> b = a >>= λ _ → b

  _<$>_ : {A B : Type} → (A → B) → M A → M B
  f <$> x = x >>= (λ a → pure (f a))

  infixl 5 _>>=_
  infixl 5 _>>_
  infixr 4 _<$>_


open Monad ⦃...⦄ public

data Maybe (A : Type) : Type where
  cake : Maybe A
  real : A → Maybe A

{-# COMPILE GHC Maybe = data Maybe (Nothing | Just) #-}

_or-else_ : {A : Type} → Maybe A → A → A
(real x) or-else _ = x
cake or-else x = x

infix 1 _or-else_

is-real : {A : Type} → Maybe A → Type
is-real cake = ⊥
is-real (real x) = ⊤

get-real : {A : Type} → (x : Maybe A) → { is-real x } → A
get-real (real x) = x

instance
  MonadMaybe : Monad Maybe
  MonadMaybe .pure = real
  (MonadMaybe ._>>=_) cake f = cake
  (MonadMaybe ._>>=_) (real x) f = f x

record Eq (T : Type) : Type₁ where
  field
    _==_ : T → T → Bool
open Eq ⦃...⦄ public

instance
  import Data.Nat
  import Data.Char
  import Data.String

  Eqℕ : Eq Data.Nat.ℕ
  Eqℕ ._==_ a b = a Data.Nat.≡ᵇ b

  EqCh : Eq Data.Char.Char
  EqCh ._==_ a b = a Data.Char.== b

  EqString : Eq String
  EqString ._==_ a b = a S.== b

  EqMaybe : {T : Type} → ⦃ Eq T ⦄ → Eq (Maybe T)
  EqMaybe ._==_ (real a) (real b) = a == b
  EqMaybe ._==_ cake     cake     = true
  EqMaybe ._==_ (real _) cake     = false
  EqMaybe ._==_ cake     (real _) = false

  EqList : {T : Type} → ⦃ Eq T ⦄ → Eq (List T)
  EqList ._==_ L.[]       L.[]       = true
  EqList ._==_ (x L.∷ xs) (y L.∷ ys) = x == y ∧ xs == ys
  EqList ._==_ (_ L.∷ _)  L.[]       = false
  EqList ._==_ L.[]       (_ L.∷ _)  = false

record Show (T : Type) : Type₁ where
  field
    show : T → String

open Show ⦃...⦄ public

instance
  open import Data.Nat.Show using () renaming (show to show-ℕ)
  ShowNat : Show ℕ
  ShowNat .show x = show-ℕ x

  ShowString : Show String
  ShowString .show x = S.show x

  ShowList : {A : Type} → ⦃ Show A ⦄ → Show (List A)
  ShowList {A} .show x = "[" S.++ go x S.++ "]"
    where
      go : List A → String
      go L.[] = ""
      go (x L.∷ L.[]) = show x
      go (x L.∷ xs) = show x S.++ ", " S.++ go xs

