open import Agda.Primitive renaming (Set to Type)

open import Data.Nat
open import Data.Bool

import Data.String as S
import Data.Char as C
import Data.List as L
open S using (String)
open C using (Char)
open L using (List)

open import Base

module HTML where

private
  _≤ᶜ_ : Char → Char → Bool
  c₀ ≤ᶜ c₁ = C.toℕ c₀ ≤ᵇ C.toℕ c₁



valid-attribute-name : String → Bool
valid-attribute-name x = L.all valid-attribute-char (S.toList x)
  where
    valid-attribute-char : Char → Bool
    valid-attribute-char c =
      ('a' ≤ᶜ c) ∧ (c ≤ᶜ 'z') ∨
      ('A' ≤ᶜ c) ∧ (c ≤ᶜ 'Z')

valid-tag-name : String → Bool
valid-tag-name x = L.all valid-tag-char (S.toList x)
  where
    valid-tag-char : Char → Bool
    valid-tag-char c =
      ('a' ≤ᶜ c) ∧ (c ≤ᶜ 'z') ∨
      ('A' ≤ᶜ c) ∧ (c ≤ᶜ 'Z') ∨
      ('0' ≤ᶜ c) ∧ (c ≤ᶜ '9')

record Attribute : Type where
  field
    name : String
    value : String
    {name-is-valid}: T (valid-attribute-name name)

instance
  open Attribute
  EqAttribute : Eq Attribute
  EqAttribute ._==_ x y = x .name == y .name ∧ x .value == y .value

record HTMLTag : Type where
  field
    name : String
    attributes : List Attribute
    {name-is-valid}: T (valid-tag-name name)

instance
  open HTMLTag

  EqTag : Eq HTMLTag
  EqTag ._==_ x y = x .name == y .name ∧ x .attributes == y .attributes

data Element : Type where
  textnode : String → Element
  node : HTMLTag → List Element → Element

{-
data HTMLTag : Type where
  html head meta title body : HTMLTag
  div p a span : HTMLTag
  h : (n : ℕ) → {T (n <ᵇ 7)} → HTMLTag

instance
  EqHTMLTag : Eq HTMLTag
  EqHTMLTag ._==_ html html = true
  EqHTMLTag ._==_ head head = true
  EqHTMLTag ._==_ meta meta = true
  EqHTMLTag ._==_ title title = true
  EqHTMLTag ._==_ body body = true
  EqHTMLTag ._==_ div div = true
  EqHTMLTag ._==_ p p = true
  EqHTMLTag ._==_ a a = true
  EqHTMLTag ._==_ span span = true
  EqHTMLTag ._==_ (h n) (h m) = n == m
  EqHTMLTag ._==_ _ _ = false

  ShowHTMLTag : Show HTMLTag
  ShowHTMLTag .show html = "html"
  ShowHTMLTag .show head = "head"
  ShowHTMLTag .show meta = "meta"
  ShowHTMLTag .show title = "title"
  ShowHTMLTag .show body = "body"
  ShowHTMLTag .show div = "div"
  ShowHTMLTag .show a = "a"
  ShowHTMLTag .show p = "p"
  ShowHTMLTag .show span = "span"
  ShowHTMLTag .show (h n) = "h" ++ˢ show n

data Attr : Type where
  lang= src= name= content= href= style= : String → Attr

attr-name : Attr → String
attr-name (lang=    _) = "lang"
attr-name (src=     _) = "src"
attr-name (name=    _) = "name"
attr-name (content= _) = "content"
attr-name (href=    _) = "href"
attr-name (style=   _) = "style"

attr-value : Attr → String
attr-value (lang=    x) = x
attr-value (src=     x) = x
attr-value (name=    x) = x
attr-value (content= x) = x
attr-value (href=    x) = x
attr-value (style=   x) = x

mutual
  DOM : Type
  DOM = List Element

  data Element : Type where
    textnode : String → Element
    element : HTMLTag → List Attr → DOM → Element

record Listable (T L : Type) : Type where
  field
    enlist : T → List L
open Listable ⦃...⦄

instance
  StringDOMAble : Listable String Element
  StringDOMAble .enlist x = textnode x ∷ []

  ListListable : {T : Type} → Listable (List T) T
  ListListable .enlist x = x

  IdListable : {T : Type} → Listable T T
  IdListable .enlist x = x ∷ []

  PairListable : {T L : Type} → ⦃ Listable T L ⦄ → Listable (L × T) L
  PairListable .enlist (x , y) = x ∷ enlist y

<_&_>_</_> : {Es As : Type} → ⦃ Listable Es Element ⦄ → ⦃ Listable As Attr ⦄ → (x : HTMLTag) → As → Es → (y : HTMLTag) → {T (x == y)} → Element
< tag & attrs > inner </ tag′ > = element tag (enlist attrs) (enlist inner)

<_>_</_> : {Es : Type} → ⦃ Listable Es Element ⦄ → (x : HTMLTag) → Es → (y : HTMLTag) → {T (x == y)} → Element
< tag > inner </ tag′ > = element tag [] (enlist inner)

<_&_/> :  {As : Type} → ⦃ Listable As Attr ⦄ → (x : HTMLTag) → As → Element
< tag & attrs /> = element tag (enlist attrs) []

<_/> : (x : HTMLTag) → Element
< tag /> = element tag [] []

encode : List Char → String → String
encode escapes text = fromList (concatMap (λ x → toList (escape-char x)) (toList text))
  where
    escape-char : Char → String
    escape-char ch =
      if any (_== ch) escapes
        then "&#" ++ˢ show (toℕ ch) ++ˢ ";"
        else fromChar ch

private
  encode-test : encode (toList "</>") "hi <3 </p>"
              ≡ "hi &#60;3 &#60;&#47;p&#62;"
  encode-test = refl

render-attr : Attr → String
render-attr x = show (attr-name x) ++ˢ "=" ++ˢ encode (toList "\t\n\r \"'=<>`") (attr-value x)

{-# TERMINATING #-} -- TODO: we should eliminate the tree as a traversable instead of doing it like this
render-element : Element → String
render-element (textnode x) = encode (toList "</>") x -- we don't technically need to escape all these characters, but it's easier to do this
render-element (element tag attrlist inner) =
  "<" ++ˢ show tag ++ˢ concatˢ (map (λ attr → " " ++ˢ render-attr attr) attrlist) ++ˢ ">"
  ++ˢ concatˢ (map render-element inner)
  ++ˢ "</" ++ˢ show tag ++ˢ ">"

sample-dom : Element
sample-dom =
  < html & lang= "en" >
    < head >
      < title > "Test page in Agda" </ title > ,
      < meta & name= "description" , content= "Simple webpage rendered in Agda DSL" /> ,
      < meta & name= "author" , content= "xenia" />
    </ head > ,
    < body >
      < p > "hello" </ p > ,
      < p > "world" </ p > ,
      < div >
        < p > "hi from within the div" </ p >
      </ div >
    </ body >
  </ html >

-}
