{-# OPTIONS --allow-unsolved-metas #-}

open import Agda.Primitive renaming (Set to Type)

open import Data.Nat
open import Data.Product

import Data.String as S
import Data.List as L
open S using (String)
open L using (List)

open import Base

open import NonEmpty
open import Bits-and-Bytes

import UTF-8
import ASCII

open import HTTP.URL

module HTTP where

module HTTP-Method where
  data Method : Type where
    GET HEAD POST PUT DELETE CONNECT OPTIONS TRACE PATCH : Method

  instance
    ShowMethod : Show Method
    ShowMethod .show x = go x
      where
        go : Method → String
        go GET     = "GET"
        go HEAD    = "HEAD"
        go POST    = "POST"
        go PUT     = "PUT"
        go DELETE  = "DELETE"
        go CONNECT = "CONNECT"
        go OPTIONS = "OPTIONS"
        go TRACE   = "TRACE"
        go PATCH   = "PATCH"

  name-of-method : Method → List⁺ Byte
  name-of-method GET     = ASCII.encode-list⁺ "GET"
  name-of-method HEAD    = ASCII.encode-list⁺ "HEAD"
  name-of-method POST    = ASCII.encode-list⁺ "POST"
  name-of-method PUT     = ASCII.encode-list⁺ "PUT"
  name-of-method DELETE  = ASCII.encode-list⁺ "DELETE"
  name-of-method CONNECT = ASCII.encode-list⁺ "CONNECT"
  name-of-method OPTIONS = ASCII.encode-list⁺ "OPTIONS"
  name-of-method TRACE   = ASCII.encode-list⁺ "TRACE"
  name-of-method PATCH   = ASCII.encode-list⁺ "PATCH"

  http-methods : List⁺ Method
  http-methods = GET ∷⁺ HEAD ∷⁺ POST ∷⁺ PUT ∷⁺ DELETE ∷⁺ CONNECT ∷⁺ OPTIONS ∷⁺ TRACE ∷⁺ [ PATCH ]⁺
open HTTP-Method public

module HTTP-Header where
  record Header : Type where
    field
      name : List⁺ Byte
      value : List⁺ Byte

  instance
    ShowHeader : Show Header
    ShowHeader .show hdr = show ⦃ ShowByteList ⦄ (list⁺-to-list (hdr .name)) S.++ ": " S.++ show ⦃ ShowByteList ⦄ (list⁺-to-list (hdr .value))
      where
        open Header


  data Content-Type : Type where
    text/plain text/html : Content-Type

  instance
    ShowContent-Type : Show Content-Type
    ShowContent-Type .show text/plain = "text/plain"
    ShowContent-Type .show text/html = "text/html"

open HTTP-Header public

module HTTP-Request where
  data RequestTargetPath : Type where
    OriginForm : Path → RequestTargetPath
    AbsoluteForm : Authority → Path → RequestTargetPath
    -- We do not care about authority-form nor asterisk form as we don't do any CONNECT nor OPTIONS requests

  record RequestTarget : Type where
    field
      path : RequestTargetPath
      query : Maybe String

  data Version : Type where
    HTTP11 : Version -- we only support 1.1

  instance
    ShowVersion : Show Version
    ShowVersion .show HTTP11 = "HTTP/1.1"

  record Request : Type where
    field
      method : Method
      target : URL
      version : Version
      headers : List Header
      content : Maybe (List Byte)
open HTTP-Request public

module HTTP-Response where
  record Response : Type where
    field
      status-code : ℕ
      status : String
      headers : List Header -- all headers besides Content-Type and Content-Length
      content : Maybe (Content-Type × List Byte) -- mime type and data

  encode-response : Response → List Byte
  encode-response res =
    UTF-8.encode-string status-line L.++ crlfᵇ L.++
    encoded-headers L.++ crlfᵇ L.++
    (proj₂ <$> res .content or-else L.[])
    where
      open Response

      crlfᵇ : List Byte
      crlfᵇ = ASCII.encode-list "\r\n"

      status-line : String
      status-line = "HTTP/1.1 " S.++ show (res .status-code) S.++ " " S.++ (res .status)

      extra-headers : List Header
      extra-headers = case res .content of λ where
        cake → L.[]
        (real (content-type , content)) →
          record
            { name = ASCII.encode-list⁺ "Content-Type"
            ; value = UTF-8.encode-string (show content-type) ++⁺ ASCII.encode-list⁺ "; charset=utf-8"
            }
          L.∷ record
            { name = ASCII.encode-list⁺ "Content-Length"
            ; value =
              list-to-list⁺? (UTF-8.encode-string (show (L.length content)))
              or-else (ASCII.encode-list⁺ "mjau")
            }
          L.∷ L.[]

      encode-header : Header → List Byte
      encode-header hdr = list⁺-to-list (hdr .name ⁺++⁺ ASCII.encode-list⁺ ": " ⁺++⁺ hdr .value)
        where
          open Header

      encoded-headers : List Byte
      encoded-headers =
        L.concat (
          L.map
            (λ h → encode-header h L.++ crlfᵇ)
            (res .headers L.++ extra-headers)
        )

  200-ok : Content-Type → String → Response
  200-ok content-type content =
    record
    { status-code = 200
    ; status = "OK"
    ; headers = L.[]
    ; content = real (content-type , UTF-8.encode-string content)
    }

  400-bad-request : Response
  400-bad-request =
    record
    { status-code = 400
    ; status = "Bad Request"
    ; headers = L.[]
    ; content = real (text/plain , ASCII.encode-list "what is bro doing")
    }

  404-not-found : Response
  404-not-found =
    record
    { status-code = 404
    ; status = "Not Found"
    ; headers = L.[]
    ; content = real (text/plain , ASCII.encode-list "not found :(")
    }

open HTTP-Response public


