open import Data.Nat
open import Data.Nat.Properties using (<ᵇ⇒<)
open import Data.Bool hiding (_<_)
open import Data.Unit
open import Data.Product

import Data.String as S
import Data.List as L
import Data.Vec as V
import Data.Char as C
open S using (String)
open L using (List)
open V using (Vec)
open C using (Char)

open import Base
open import NonEmpty
open import Bits-and-Bytes
open import Indexed
open import Parsing (Byte)

open import HTTP
open import ParseHelpers

import UTF-8
import ASCII

module HTTP.Parse where

-- helper functions
private
  decode-utf8⁺ : List⁺ Byte → Maybe String
  decode-utf8⁺ bs = UTF-8.decode-string (list⁺-to-list bs)

module Parse-URL where
  open import HTTP.URL

  -- commonly used sub-parsers
  private
    gen-delims sub-delims reserved unreserved : [ Parser Byte ]
    gen-delims = any-of ":/?#[]@"
    sub-delims = any-of "!$&'()*+,;="
    reserved = gen-delims <|>′ sub-delims
    unreserved = letter <|>′ digit <|>′ any-of "-._~"

    pchar : [ Parser Byte ]
    pchar = unreserved <|>′ percent-encoded <|>′ sub-delims <|>′ any-of ":@"

  parse-scheme : [ Parser String ]
  parse-scheme = decode-utf8⁺ <$?>′ many (letter <|>′ digit <|>′ any-of "+-.")

  parse-authority : [ Parser Authority ]
  parse-authority = (λ auth → auth .port <ᵇ 2 ^ 16) <?>′ unchecked
    where
      reg-name : [ Parser String ]
      reg-name = decode-utf8⁺ <$?>′ (many (unreserved <|>′ percent-encoded <|>′ sub-delims))

      unchecked : [ Parser Authority ]
      unchecked = (λ (host , rest) → host ꞉ ((proj₂ <$> rest) or-else 80))
                  <$>′ (reg-name <&?>□ any-of ":" <&>□ number)

  parse-path : [ Parser Path ]
  parse-path =
      L.foldr (λ (slash , seg) path → seg / path) $ <$>′
      (
        (list⁺-to-list <$>′ many (any-of "/" <&>′ enbox segment))
        <|>′ ((λ _ → L.[]) <$>′ any-of "/")
      )
    where
      segment : [ Parser String ]
      segment = decode-utf8⁺ <$?>′ many pchar

  parse-query parse-fragment : [ Parser String ]
  parse-query = decode-utf8⁺ <$?>′ many (pchar <|>′ any-of "/?")
  parse-fragment = parse-query

  parse-absolute-url : [ Parser (Authority × Path) ]
  parse-absolute-url =
    (parse-scheme >?=′ (_== "http"))
    <&⃗>□ exact (ASCII.encode-vec "://")
    <&⃗>□ parse-authority
    <&>□ parse-path

  parse-url : [ Parser URL ]
  parse-url =
    parse-absolute-url
    <&?>□ (any-of "?" <&⃗>□ parse-query)
    >$=′ λ ((auth , path) , mquery) → http:// auth / path ¿ mquery

open Parse-URL public

module Parse-Header where
  open import HTTP.URL using (Authority)
  open HTTP.Header

  get-field : List⁺ Byte → List Header → Maybe (List⁺ Byte)
  get-field field-name L.[] = cake
  get-field field-name (hdr L.∷ rest) =
    if hdr .name == field-name
      then real (hdr .value)
      else get-field field-name rest

  content-length : List Header → Maybe ℕ
  content-length headers =
    get-field (ASCII.encode-list⁺ "Content-Length") headers
    >>= parse-list⁺ number

  host : List Header → Maybe Authority
  host headers =
    get-field (ASCII.encode-list⁺ "Host") headers
    >>= parse-list⁺ parse-authority

  private
    parse-vchar-sp : [ Parser Byte ]
    parse-vchar-sp = (λ x → (0x20 ≤ᵇ x .value) ∧ (x .value ≤ᵇ 0x7e)) <?>′ any₁

    parse-token : [ Parser (List⁺ Byte) ]
    parse-token = many parse-tchar
      where
        parse-tchar : [ Parser Byte ]
        parse-tchar = digit <|>′ letter <|>′ any-of "!#$%&'*+-.^_`|~"

  parse-field-content : [ Parser (List⁺ Byte) ]
  parse-field-content = many parse-vchar-sp

  parse-header : [ Parser Header ]
  parse-header =
    (λ (name , value) → record { name = name ; value = value })
    <$>′ (
      parse-token
      <&>□ any-of ":"
      <&⃗>□ spaces
      <&⃗>□ parse-field-content
    )

  parse-headers⁺ : [ Parser (List⁺ Header) ]
  parse-headers⁺ = many parse-header

open Parse-Header public

module Parse-Request where
  open import HTTP.URL

  parse-request-target : [ Parser RequestTarget ]
  parse-request-target =
    (parse-origin-form <|>′ parse-absolute-form) <&?>□ (any-of "?" <&⃗>□ parse-query)
    >$=′ λ (path , query) → record { path = path; query = query }
    where
      parse-origin-form parse-absolute-form : [ Parser RequestTargetPath ]
      parse-origin-form =
        parse-path >$=′ OriginForm

      parse-absolute-form =
        parse-absolute-url >$=′ λ (auth , path) → AbsoluteForm auth path

  parse-http-version : [ Parser Version ]
  parse-http-version = (λ _ → HTTP11) <$>′ exact (ASCII.encode-vec "HTTP/1.1")

  parse-specific-http-method : Method → [ Parser Method ]
  parse-specific-http-method m = (λ _ → m) <$>′ exact (proj₂ (list⁺-to-vec (name-of-method m)))

  parse-http-method : [ Parser Method ]
  parse-http-method = foldr (λ m p → parse-specific-http-method m <|>′ p) fail http-methods

  parse-request-line : [ Parser (Method × RequestTarget × Version) ]
  parse-request-line =
    parse-http-method
    <&>□ (
      spaces
      <&⃗>□ (
        parse-request-target
        <&>□ (
          spaces
          <&⃗>□ parse-http-version
        )
      )
    )



  parse-request : [ Parser Request ]
  parse-request =
    ((parse-request-line <&⃖>□ crlf) <&>□ many (parse-header <&⃖>□ crlf))
    >$?=′ find-target
    >>=′ λ (method , target , version , headers) → enbox (
      case content-length (list⁺-to-list headers) of λ where
        cake →
          crlf >$=′ λ _ →
            record
              { method = method
              ; target = target
              ; version = version
              ; headers = list⁺-to-list headers
              ; content = cake
              }
        (real 0) →
          (crlf <&>□ crlf) >$=′ λ _ →
            record
              { method = method
              ; target = target
              ; version = version
              ; headers = list⁺-to-list headers
              ; content = real L.[]
              }
        (real n@(suc _)) →
          (crlf <&⃗>□ repeat n any₁) >$=′ λ content →
            record
              { method = method
              ; target = target
              ; version = version
              ; headers = list⁺-to-list headers
              ; content = real (V.toList content)
              }
    )
    where
      find-target : ((Method × RequestTarget × Version) × List⁺ Header) → Maybe (Method × URL × Version × List⁺ Header)
      find-target ((method , record { path = OriginForm path ; query = query } , ver) , headers) =
        (λ auth → method , (http:// auth / path ¿ query) , ver , headers) <$> Parse-Header.host (list⁺-to-list headers)
      find-target ((method , record { path = AbsoluteForm auth path ; query = query } , ver) , headers) =
        real (method , (http:// auth / path ¿ query) , ver , headers)

open Parse-Request public
